/*
 * Copyright (c) 1997-2003 Erez Zadok
 * Copyright (c) 2001-2003 Stony Brook University
 * Copyright (c) 1997-2000 Columbia University
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: inode.c,v 1.19 2002/12/27 20:19:04 ezk Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#ifdef FISTGEN
# include "fist_logeventfs.h"
#endif /* FISTGEN */
#include "fist.h"
#include "logeventfs.h"


STATIC int
logeventfs_create(inode_t *dir, dentry_t *dentry, int mode)
{
    int err;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;
    const char *name;

    print_entry_location();
    hidden_dentry = logeventfs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    name = dentry->d_name.name;
    ASSERT(hidden_dentry != NULL);
    fist_checkinode(dir, "logeventfs_create");

    ;

    hidden_dir_dentry = lock_parent(hidden_dentry);
    err = PTR_ERR(hidden_dir_dentry);
    if (IS_ERR(hidden_dir_dentry))
	goto out;

    err = vfs_create(hidden_dir_dentry->d_inode, hidden_dentry, mode);
    /* XXX this could potentially return a negative hidden_dentry! */
    if (err)
	goto out_lock;

    /* precall and postcall added - hooked on l24. */
/* test with "touch {newfile}". */



 {					
	printk("logeventfs: create:   file %s\n", name);
};

    err = logeventfs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (err)
	goto out_lock;

    fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);


 out_lock:
    unlock_dir(hidden_dir_dentry);
 out:
    fist_checkinode(dir, "post logeventfs_create");
    print_exit_status(err);
    return err;
}


STATIC dentry_t *
logeventfs_lookup(inode_t *dir, dentry_t *dentry)
{
    int err = 0;
    dentry_t *hidden_dir_dentry;
    dentry_t *hidden_dentry;
    const char *name;
    vnode_t *this_vnode;
    dentry_t *this_dir;
    unsigned int namelen;

    print_entry_location();
    hidden_dir_dentry = logeventfs_hidden_dentry(dentry->d_parent); /* CPW: Moved below print_entry_location */
    name = dentry->d_name.name;
    namelen = dentry->d_name.len;
    fist_checkinode(dir, "logeventfs_lookup");

    this_vnode = dir;
    this_dir = hidden_dir_dentry;

    fist_print_dentry("logeventfs_lookup IN", dentry);
    fist_print_dentry("logeventfs_lookup: dentry->d_parent IN", dentry->d_parent);
    fist_print_dentry("logeventfs_lookup: hidden_dir_dentry IN", hidden_dir_dentry);
    fist_print_inode("logeventfs_lookup: dir IN", dir);

    if (hidden_dir_dentry->d_inode)
	fist_print_inode("logeventfs_lookup: hidden_dir_dentry->d_inode",
			 hidden_dir_dentry->d_inode);

    /* must initialize dentry operations */
    dentry->d_op = &logeventfs_dops;

    ;

    /* increase refcount of base dentry (lookup_one[_len] will decrement) */
    // THIS IS RIGHT! (don't "fix" it)
    // NO THIS IS WRONG IN 2.3.99-pre6. lookup_one[_len] will NOT decrement
    // dget(hidden_dir_dentry);

    /* will allocate a new hidden dentry if needed */
    hidden_dentry = lookup_one_len(name, hidden_dir_dentry, namelen);

    if (IS_ERR(hidden_dentry)) {
	/*
	 * this produces an unusual dentry: one that has neither an
	 * inode, nor a private structure attached to it. All cleanup
	 * methods (d_delete, d_release, etc) must be prepared to deal
	 * with such dentries. Ion 09/29/2001
	 */
	printk("ERR from hidden_dentry!!!\n");
	err = PTR_ERR(hidden_dentry);
	goto out;
    }

    /* Always had precall and postcall - hooked on l24. */
/* test with "stat {nonexistant_file}". */



 {
	printk("logeventfs: lookup:   file %s\n", name);
};

    /* update parent directory's atime */
    fist_copy_attr_atime(dir, hidden_dir_dentry->d_inode);
    /* link the upper and lower dentries */
    dtopd(dentry) = (struct logeventfs_dentry_info *) kmalloc(sizeof(struct logeventfs_dentry_info), GFP_KERNEL);
    if (!dtopd(dentry)) {
	err = -ENOMEM;
	goto out_dput;
    }
    dtohd(dentry) = hidden_dentry;


    /* lookup is special: it needs to handle negative dentries */
    if (!hidden_dentry->d_inode) {
	d_add(dentry, NULL);
	fist_print_dentry("logeventfs_lookup OUT hidden_dentry", hidden_dentry);
	goto out;
    }

    fist_dprint(6, "lookup \"%s\" -> inode %d\n", name, hidden_dentry->d_inode->i_ino);
    err = logeventfs_interpose(hidden_dentry, dentry, dir->i_sb, 1);
    if (err)
	goto out_free;

    fist_checkinode(dentry->d_inode, "logeventfs_lookup OUT: dentry->d_inode:");
    fist_checkinode(dir, "logeventfs_lookup OUT: dir");

    fist_print_dentry("logeventfs_lookup OUT hidden_dentry", hidden_dentry);
    fist_print_inode("logeventfs_lookup OUT hidden_inode", hidden_dentry->d_inode);


    /* All is well */
    goto out;

 out_free:
    d_drop(dentry);		/* so that our bad dentry will get destroyed */
    kfree(dtopd(dentry));
    dtopd(dentry) = NULL;	/* be safe */

 out_dput:
    dput(hidden_dentry);

 out:
    fist_print_dentry("logeventfs_lookup OUT", dentry);
    print_exit_status(err);
    return ERR_PTR(err);
}


STATIC int
logeventfs_link(dentry_t *old_dentry, inode_t *dir, dentry_t *new_dentry)
{
    int err;
    dentry_t *hidden_old_dentry;
    dentry_t *hidden_new_dentry;
    dentry_t *hidden_dir_dentry;
    /* WLS: How to handle the fact that link has an old and new name? */
    const char *name;

    print_entry_location();
    hidden_old_dentry = logeventfs_hidden_dentry(old_dentry); /* CPW: Moved below print_entry_location */
    hidden_new_dentry = logeventfs_hidden_dentry(new_dentry);
    name = new_dentry->d_name.name;	/* WLS Likewise, how to handle other dentry? */

    fist_checkinode(dir, "logeventfs_link-dir");
    fist_checkinode(old_dentry->d_inode, "logeventfs_link-oldinode");

    ;

    dget(hidden_old_dentry);
    dget(hidden_new_dentry);
    hidden_dir_dentry = lock_parent(hidden_new_dentry);


    err = vfs_link(hidden_old_dentry,
		   hidden_dir_dentry->d_inode,
		   hidden_new_dentry);
    if (err || !hidden_new_dentry->d_inode)
	goto out_lock;

    /* precall and postcall added - hooked on l24. */
/* test with "ln -f {file} {newname}". */



 {
	printk("logeventfs: link:     file %s\n", name);
};

    err = logeventfs_interpose(hidden_new_dentry, new_dentry, dir->i_sb, 0);
    if (err)
	goto out_lock;

    fist_copy_attr_timesizes(dir, hidden_new_dentry->d_inode);
    /* propagate number of hard-links */
    old_dentry->d_inode->i_nlink = itohi(old_dentry->d_inode)->i_nlink;

 out_lock:
    unlock_dir(hidden_dir_dentry);
    dput(hidden_new_dentry);
    dput(hidden_old_dentry);
    if (!new_dentry->d_inode)
	d_drop(new_dentry);

    print_exit_status(err);
    return err;
}


STATIC int
logeventfs_unlink(inode_t *dir, dentry_t *dentry)
{
    int err = 0;
    inode_t *hidden_dir; 
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;
    const char *name;

    print_entry_location();

    hidden_dir = itohi(dir); /* CPW: Moved below print_entry_location */
    hidden_dentry = logeventfs_hidden_dentry(dentry);
    name = dentry->d_name.name;

    ASSERT(hidden_dentry != NULL);
    fist_checkinode(dir, "logeventfs_unlink-dir");

    dget(dentry);
    hidden_dir_dentry = lock_parent(hidden_dentry);

    ;


    /* avoid destroying the hidden inode if the file is in use */
    dget(hidden_dentry);
    err = vfs_unlink(hidden_dir, hidden_dentry);
    dput(hidden_dentry);

    if (!err)			/* vfs_unlink does that */
	d_delete(hidden_dentry);

    /* precall and postcall added - hooked on l24. */
/* test with "rm -f {file}". */



 {
	printk("logeventfs: unlink:   file %s\n", name);
};

 out_lock:
    fist_copy_attr_times(dir, hidden_dir);
    /* propagate number of hard-links */
    dentry->d_inode->i_nlink = itohi(dentry->d_inode)->i_nlink;

    unlock_dir(hidden_dir_dentry);

    /*
     * call d_drop so the system "forgets" about us
     */
    if (!err) {
	//	fist_print_dentry("logeventfs_unlink: 0", dentry);
	d_drop(dentry);
	//	fist_print_dentry("logeventfs_unlink: 1", dentry);
    }

    dput(dentry);

    fist_checkinode(dir, "post logeventfs_unlink-dir");
    print_exit_status(err);
    return err;
}


STATIC int
logeventfs_symlink(inode_t *dir, dentry_t *dentry, const char *symname)
{
    int err;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;
    const char *name;

    print_entry_location();
    hidden_dentry = logeventfs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    name = dentry->d_name.name;
    fist_checkinode(dir, "logeventfs_symlink-dir");

    dget(hidden_dentry);
    hidden_dir_dentry = lock_parent(hidden_dentry);

    ;

    err = vfs_symlink(hidden_dir_dentry->d_inode,
		      hidden_dentry,
		      symname);
    if (err || !hidden_dentry->d_inode)
	goto out_lock;
    err = logeventfs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (err)
	goto out_lock;

    /* precall and postcall added - hooked on l24. */
/* test with "ln -s {file} {newname}". */



 {
	printk("logeventfs: symlink:  file %s\n", name);
};

    fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);
    fist_checkinode(dir, "post logeventfs_symlink-dir");

 out_lock:
    unlock_dir(hidden_dir_dentry);
    dput(hidden_dentry);
    if (!dentry->d_inode)
	d_drop(dentry);

    print_exit_status(err);
    return err;
}


STATIC int
logeventfs_mkdir(inode_t *dir, dentry_t *dentry, int mode)
{
    int err;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;
    const char *name;

    print_entry_location();
    hidden_dentry = logeventfs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    name = dentry->d_name.name;
    fist_checkinode(dir, "logeventfs_mkdir-dir");

    hidden_dir_dentry = lock_parent(hidden_dentry);

    ;

    err = vfs_mkdir(hidden_dir_dentry->d_inode,
		    hidden_dentry,
		    mode);
    if (err || !hidden_dentry->d_inode)
	goto out;

    err = logeventfs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (err)
	goto out;

    ;

    fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);
    /* update number of links on parent directory */
    dir->i_nlink = hidden_dir_dentry->d_inode->i_nlink;

    fist_checkinode(dir, "post logeventfs_mkdir-dir");

 out:
    unlock_dir(hidden_dir_dentry);
    if (!dentry->d_inode)
	d_drop(dentry);

    print_exit_status(err);
    return err;
}


STATIC int
logeventfs_rmdir(inode_t *dir, dentry_t *dentry)
{
    int err = 0;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;
    const char *name;

    print_entry_location();
    hidden_dentry = logeventfs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    name = dentry->d_name.name;
    fist_checkinode(dir, "logeventfs_rmdir-dir");

    dget(dentry);
    hidden_dir_dentry = lock_parent(hidden_dentry);

    ;

    /* avoid destroying the hidden inode if the file is in use */
    dget(hidden_dentry);
    err = vfs_rmdir(hidden_dir_dentry->d_inode, hidden_dentry);
    dput(hidden_dentry);

    if (!err)			/* vfs_rmdir does that */
	d_delete(hidden_dentry);

    /* precall and postcall added - hooked on l24. */
/* test with "rmdir {empty_dir}". */



 {
	printk("logeventfs: rmdir:    directory %s\n", name);
};

 out_lock:
    fist_copy_attr_times(dir, hidden_dir_dentry->d_inode);
    /* copy the nlink count for our dentry and our parent's dentry */
    dir->i_nlink =  hidden_dir_dentry->d_inode->i_nlink;

    unlock_dir(hidden_dir_dentry);

    /*
     * call d_drop so the system "forgets" about us
     */
    if (!err)
	d_drop(dentry);

    dput(dentry);

    print_exit_status(err);
    return err;
}


STATIC int
logeventfs_mknod(inode_t *dir, dentry_t *dentry, int mode, int dev)
{
    int err;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;

    print_entry_location();
    hidden_dentry = logeventfs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    fist_checkinode(dir, "logeventfs_mknod-dir");

    hidden_dir_dentry = lock_parent(hidden_dentry);

    err = vfs_mknod(hidden_dir_dentry->d_inode,
		    hidden_dentry,
		    mode,
		    dev);
    if (err || !hidden_dentry->d_inode)
	goto out;

    err = logeventfs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (err)
	goto out;
    fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);

 out:
    unlock_dir(hidden_dir_dentry);
    if (!dentry->d_inode)
	d_drop(dentry);

    fist_checkinode(dir, "post logeventfs_mknod-dir");
    print_exit_status(err);
    return err;
}


STATIC int
logeventfs_rename(inode_t *old_dir, dentry_t *old_dentry,
	      inode_t *new_dir, dentry_t *new_dentry)
{
    int err;
    dentry_t *hidden_old_dentry;
    dentry_t *hidden_new_dentry;
    dentry_t *hidden_old_dir_dentry;
    dentry_t *hidden_new_dir_dentry;
    const char *name;

    print_entry_location();

    hidden_old_dentry = logeventfs_hidden_dentry(old_dentry);/* CPW: Moved below print_entry_location */
    hidden_new_dentry = logeventfs_hidden_dentry(new_dentry);
    name = old_dentry->d_name.name;		/* WLS How to handle two dentry function? */

    fist_checkinode(old_dir, "logeventfs_rename-old_dir");
    fist_checkinode(new_dir, "logeventfs_rename-new_dir");

    ;

    dget(hidden_old_dentry);
    dget(hidden_new_dentry);
    hidden_old_dir_dentry = get_parent(hidden_old_dentry);
    hidden_new_dir_dentry = get_parent(hidden_new_dentry);
    double_lock(hidden_old_dir_dentry, hidden_new_dir_dentry);


    err = vfs_rename(hidden_old_dir_dentry->d_inode, hidden_old_dentry,
		     hidden_new_dir_dentry->d_inode, hidden_new_dentry);
    if (err)
	goto out_lock;

    /* precall and postcall added - hooked on l24. */
/* test with "mv {file} {newname}". */




 {
	/* printk("logeventfs: rename:   from %s/%s to %s/%s\n", $fromdir, $from, $todir, $to); */
	printk("logeventfs: rename:   from %s\n", name);
};

    fist_copy_attr_all(new_dir, hidden_new_dir_dentry->d_inode);
    if (new_dir != old_dir)
	fist_copy_attr_all(old_dir, hidden_old_dir_dentry->d_inode);

 out_lock:
    // double_unlock will dput the new/old parent dentries whose refcnts
    // were incremented via get_parent above.
    double_unlock(hidden_old_dir_dentry, hidden_new_dir_dentry);
    dput(hidden_new_dentry);
    dput(hidden_old_dentry);

    fist_checkinode(new_dir, "post logeventfs_rename-new_dir");
    print_exit_status(err);
    return err;
}


STATIC int
logeventfs_readlink(dentry_t *dentry, char *buf, int bufsiz)
{
    int err;
    dentry_t *hidden_dentry;
    const char *name;

    print_entry_location();
    hidden_dentry = logeventfs_hidden_dentry(dentry);/* CPW: Moved below print_entry_location */
    name = dentry->d_name.name;
    fist_print_dentry("logeventfs_readlink IN", dentry);

    if (!hidden_dentry->d_inode->i_op ||
	!hidden_dentry->d_inode->i_op->readlink) {
	err = -EINVAL;
	goto out;
    }

    ;

    err = hidden_dentry->d_inode->i_op->readlink(hidden_dentry,
						 buf,
						 bufsiz);
    if (err > 0)
	fist_copy_attr_atime(dentry->d_inode, hidden_dentry->d_inode);

    /* precall and postcall added - hooked on l24. */
/* test with "ls -al {dir_with_a_symlink}". */



 {
	printk("logeventfs: readlink: file %s\n", name);
};

 out:
    print_exit_status(err);
    return err;
}


STATIC int
logeventfs_follow_link(dentry_t *dentry, struct nameidata *nd)
{
    char *buf;
    int len = PAGE_SIZE, err;
    mm_segment_t old_fs;

    print_entry_location();
    //    fist_print_dentry("logeventfs_follow_link dentry IN", dentry);

    buf = kmalloc(len, GFP_KERNEL);
    if (!buf) {
	err = -ENOMEM;
	goto out;
    }

    /* read the symlink, and then we will follow it */
    old_fs = get_fs();
    set_fs(KERNEL_DS);
    err = dentry->d_inode->i_op->readlink(dentry, buf, len);
    set_fs(old_fs);
    if (err < 0)
	goto out_free;

    buf[err] = 0;	// terminate the buffer -- XXX still needed?

    // XXX: FIXME w/ logeventfs_encode_filename()
    /*
     * vfs_follow_link will increase the nd's mnt refcnt
     * we assume that some other VFS code decrements it.
     */
    err = vfs_follow_link(nd, buf);

 out_free:
    kfree(buf);
 out:
#if 0
    if (err < 0) {
	dput(nd->dentry);
	printk("EZK follow_link() mnt_cnt %d\n", atomic_read(&nd->mnt->mnt_count));
	mntput(nd->mnt);
    }
#endif

    print_exit_status(err);
    return err;
}




STATIC int
logeventfs_permission(inode_t *inode, int mask)
{
    inode_t *hidden_inode;
    int mode;
    int err;

    print_entry_location();
    hidden_inode = itohi(inode);/* CPW: Moved below print_entry_location */
    mode = inode->i_mode;

    err = vfs_permission(inode, mask);
    if (err)
	goto out;
    err = permission(hidden_inode, mask);
 out:
    print_exit_status(err);
    return err;
}


STATIC int
logeventfs_inode_revalidate(dentry_t *dentry)
{
    int err = 0;
    dentry_t *hidden_dentry;
    inode_t *hidden_inode;

    print_entry_location();
    hidden_dentry = logeventfs_hidden_dentry(dentry);/* CPW: Moved below print_entry_location */
    hidden_inode = hidden_dentry->d_inode;
    //    fist_print_dentry("logeventfs_inode_revalidate IN", dentry);

    if (hidden_inode->i_op && hidden_inode->i_op->revalidate) {
	err = hidden_inode->i_op->revalidate(hidden_dentry);
	if (!err)
	    fist_copy_attr_all(dentry->d_inode, hidden_inode);
    }

    //    fist_print_dentry("logeventfs_inode_revalidate OUT", dentry);
    print_exit_status(err);
    return err;
}


STATIC int
logeventfs_setattr(dentry_t *dentry, struct iattr *ia)
{
    int err = 0;
    dentry_t *hidden_dentry;
    const char *name;
    inode_t *inode;
    inode_t *hidden_inode;

    print_entry_location();
    hidden_dentry = logeventfs_hidden_dentry(dentry);
    name = dentry->d_name.name;
    inode = dentry->d_inode;
    hidden_inode = itohi(inode);
    fist_checkinode(inode, "logeventfs_setattr");

    ;

    err = notify_change(hidden_dentry, ia);

    /* precall and postcall added - hooked on l24. */
/* test with "touch". */



 {
	printk("logeventfs: setattr:  file %s\n", name);
};

#if defined(FIST_FILTER_DATA) || defined(FIST_FILTER_SCA)
 out:
#endif /* FIST_FILTER_DATA || FIST_FILTER_SCA */
    /*
     * The lower file system might has changed the attributes, even if
     * notify_change above resulted in an error(!)  so we copy the
     * hidden_inode's attributes (and a few more) to our inode.
     */
    fist_copy_attr_all(inode, hidden_inode);

    fist_checkinode(inode, "post logeventfs_setattr");
    print_exit_status(err);
    return err;
}


#if NOT_USED_YET
STATIC int
logeventfs_getattr(dentry_t *dentry, struct iattr *ia)
{
    /*
     * Kinda pointless at the moment as logeventfs_getattr's disabled, but...
     */
    const char *name;

    name = dentry->d_name.name;

    ;



    /* wrapfs_getattr is not enabled in the linux24 template at all, but precall and postcall added anyways. */
/* test with "". */



 {
	printk("logeventfs: getattr:  file %s\n", name);
};

    return -ENOSYS;
}
#endif /* NOT_USED_YET */


struct inode_operations logeventfs_symlink_iops =
{
    readlink:	logeventfs_readlink,
    follow_link: logeventfs_follow_link,
    permission:	logeventfs_permission,
    revalidate:	logeventfs_inode_revalidate,
    setattr:	logeventfs_setattr,
#if 0
    // XXX: off, b/c the VFS doesn't call getattr yet
    getattr:	logeventfs_getattr,
#endif
};

struct inode_operations logeventfs_dir_iops =
{
    create:	logeventfs_create,
    lookup:	logeventfs_lookup,
    link:	logeventfs_link,
    unlink:	logeventfs_unlink,
    symlink:	logeventfs_symlink,
    mkdir:	logeventfs_mkdir,
    rmdir:	logeventfs_rmdir,
    mknod:	logeventfs_mknod,
    rename:	logeventfs_rename,
    /* no readlink/follow_link for non-symlinks */
    // off because we have setattr
    //    truncate:	logeventfs_truncate,
    permission:	logeventfs_permission,
    revalidate:	logeventfs_inode_revalidate,
    setattr:	logeventfs_setattr,
#if 0
    // XXX: off, b/c the VFS doesn't call getattr yet
    getattr:	logeventfs_getattr,
#endif
};

struct inode_operations logeventfs_main_iops =
{
    permission:	logeventfs_permission,
    revalidate:	logeventfs_inode_revalidate,
    setattr:	logeventfs_setattr,
#if 0
    // XXX: off, b/c the VFS doesn't call getattr yet
    getattr:	logeventfs_getattr,
#endif
};

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
